open MyUtils
  
open CompAst

module SI = SmtZ3Interface

(* Package: AstSimulator. Utilities to simulate the AST using a concolic simulation *)
type cVal = | FloatVal of float | EnumVal of int | BoolVal of bool

type stateType =
    { paramValues : cVal Inthash.t;
    (*-- values of the parameters, chosen at start of the simulation --*)
      varValues : cVal Inthash.t;
    (*-- values of variables at current depth. --*)
      clockValues : (bool * float) Inthash.t
    }

let toString cVal = 
  match cVal with 
    | BoolVal b -> if (b) then "true" else "false"
    | FloatVal f -> string_of_float f
    | EnumVal i -> string_of_int i

let convertConstant v =
  match v with
    | CompAst.BoolConst b -> BoolVal b
    | CompAst.NumConst f -> FloatVal f
    | CompAst.EnumConst i -> EnumVal i
    | _ -> assert false


let encodeConstant ctx cval = 
  match cval with 
    | BoolVal b -> 
      SI.boolConstant ctx b
    | FloatVal f -> 
      SI.floatConstant ctx f
    | EnumVal i -> 
      SI.intConstant ctx i

let encodeParameterValues pVals encInfo = 
  let handleParamValue pID cVal = 
    let cExpr = encodeConstant  encInfo.SmtEncoder.ctx cVal in 
    Inthash.replace encInfo.SmtEncoder.pHash pID cExpr
  in 
  Inthash.iter handleParamValue pVals


let boolOperation bO b1 b2 = 
  match (b1,b2) with 
    | (BoolVal b1Val), (BoolVal b2Val) ->  (
      match bO with 
	| CompAst.AAnd -> (BoolVal (b1Val && b2Val))
	| CompAst.AOr -> (BoolVal (b1Val || b2Val))
	| CompAst.AImpl -> (BoolVal (( not b1Val) || b2Val))
    )
    | _ -> assert false (* This cannot happen *)

let boolNot b = 
  match b with 
    | BoolVal bVal -> (BoolVal (not bVal))
    | _ -> assert false 


let relOperation rO e1 e2 = 
  match (e1,e2) with 
    | (FloatVal f1), (FloatVal f2) -> 
      (  match rO with 
	| CompAst.AGeq -> BoolVal (f1 >= f2)
	| CompAst.AGt -> BoolVal (f1 > f2)
	| CompAst.AEq -> BoolVal (f1 = f2)
	| CompAst.ANeq -> BoolVal ( not (f1 = f2))
      )
    | (EnumVal f1), (EnumVal f2) -> (
      match rO with 
	| CompAst.AGeq -> BoolVal (f1 >= f2)
	| CompAst.AGt -> BoolVal (f1 > f2)
	| CompAst.AEq -> BoolVal (f1 = f2)
	| CompAst.ANeq -> BoolVal ( not (f1 = f2) )		
    )
    | (BoolVal b1), (BoolVal b2) -> (
      match rO with 
	| CompAst.AGeq 
	| CompAst.AGt -> assert false
	| CompAst.AEq -> BoolVal (b1 = b2)
	| CompAst.ANeq -> BoolVal ( not (b1 = b2)	)	
    ) 
    | _ -> assert false 

let arithOperation bO e1 e2 = 
  match (e1,e2) with 
    | (FloatVal f1), (FloatVal f2) -> (
      match bO with 
	| CompAst.APlus -> (FloatVal (f1 +. f2))
	| CompAst.AMinus -> (FloatVal (f1 -. f2))
	| CompAst.AStar -> (FloatVal (f1 *. f2))
	| CompAst.ADiv -> (FloatVal (f1 /. f2))
    )
    | _ -> assert false 
      
let negOperation e = 
  match e with 
    | FloatVal f -> FloatVal (-1.0 *. f)
    | _ -> assert false
      
let getValue (cState,ndHash) vi = 
  if (vi.CompAst.isNondet) 
  then ( try Inthash.find ndHash vi.CompAst.vID with | Not_found -> assert false)
  else (
    if (vi.CompAst.isClock)
    then (
      let (b,f) = try Inthash.find cState.clockValues vi.CompAst.vID with | Not_found -> assert false 
      in FloatVal f
    ) else (
      try Inthash.find cState.varValues vi.CompAst.vID with | Not_found -> assert false
    )	
  )


let rec evalExpr cState expr = 
  match expr with 
    |  CompAst.ABoolOp (bO,e1,e2) -> (  let b1 = evalExpr cState e1 in 
					let b2 = evalExpr cState e2 in 
					boolOperation bO b1 b2 ) 
    | CompAst.ANot e -> ( let b = evalExpr cState e in boolNot b )
    | CompAst.ARelOp(rO,e1,e2) -> ( let e1Val = evalExpr cState e1 in 
				    let e2Val = evalExpr cState e2 in 
				    relOperation rO e1Val e2Val )
      
    | CompAst.AArithOp(bO,e1,e2) -> ( let e1Val = evalExpr cState e1 in 
				      let e2Val = evalExpr cState e2 in 
				      arithOperation bO e1Val e2Val )
    | CompAst.ANeg e -> ( let b = evalExpr cState e in negOperation b )
    | CompAst.AVar vi -> (getValue cState  vi)
    | CompAst.AConst cI -> convertConstant cI  
    | CompAst.ADefined pI ->  BoolVal true
      



(*-- values of clocks that include whether the clock is running and what its value is --*)
let printState oc gInfo cst =
  let printValue str cVal =
    match cVal with
      | FloatVal f -> Printf.fprintf oc "%s = %f, " str f
      | EnumVal i ->
        Printf.fprintf oc "%s = %s, " str (CompAst.string_of_enum_id gInfo i)
      | BoolVal b ->
        Printf.fprintf oc "%s = %s, " str (if b then "true" else "false") in
  let hParams pId cval =
    let pInfo = CompAst.getParamInfoFromID gInfo pId
    in printValue pInfo.CompAst.pName cval in
  let hVars vId cVal =
    let vName = CompAst.getVariableNameFromID gInfo vId
    in printValue vName cVal in
  let hClock cId (b, f) =
    let vInfo = CompAst.getVariableNameFromID gInfo cId
    in
    (printValue vInfo (FloatVal f);
     printValue (vInfo ^ "_is_running") (BoolVal b))
  in
  (Inthash.iter hParams cst.paramValues;
   Inthash.iter hVars cst.varValues;
   Inthash.iter hClock cst.clockValues)

    
(* Not possible *)
let getParamValues gInfo =
  let pHash = Inthash.create 100 in
  let handleParam pid pI =
    let v = CompAst.chooseInitialParamValueAtRandom pI in
    let vval = convertConstant v in Inthash.add pHash pid vval
  in (Inthash.iter handleParam gInfo.CompAst.gParams; pHash)
  
let printParamVals gInfo pVals = 
  let hParam pid cVal = 
    let pinfo = CompAst.getParamInfoFromID gInfo pid in 
    let cStr = toString cVal in 
    Printf.printf "%s  %s \n" pinfo.CompAst.pName cStr
  in 
  Inthash.iter hParam pVals


let getVariableAndClockValues paramVals gInfo =
  (* 1. Create an encoding info object for invoking the solver.*)
  (* 2. Add all expressions for initial conditions. *)
  (* 3. Extract a model *)
  let encInfo =
    {
      SmtEncoder.ctx = SI.createContext ();
      SmtEncoder.pHash = Inthash.create 100;
      SmtEncoder.vHash = IntPairhash.create 100;
      SmtEncoder.vClockFrozenHash = IntPairhash.create 10;
      SmtEncoder.vTrsTakenHash = IntPairhash.create 10;
      SmtEncoder.curDepth = 0;
    }
  in
    (* A hashtable from the parameters to the variables *)
    (* A hashtable from the variables at particular depth to the corr. variables *)
    (* Variable to track whether a clock is frozen or not *)
    (* transition ID, depth -> encoding *)
  ( encodeParameterValues  paramVals encInfo;
    SmtEncoder.createVariablesAtCurrentDepth gInfo encInfo;
    SmtEncoder.assertInitialConditions gInfo encInfo;
   (match SI.check encInfo.SmtEncoder.ctx with
     | SI.SAT -> ()
     | SI.FAIL 
     | SI.UNSAT -> 
       (
	 Printf.fprintf stderr "Fatal: Could not find suitable initial conditions with parameter values \n";
	 printParamVals gInfo paramVals;
	 assert false
       )
   );
   let sModel = SI.getModel encInfo.SmtEncoder.ctx in
     (* We can now extract state for clock and other variables *)
   let vHash = Inthash.create 100 in
   let cHash = Inthash.create 100 in
   let hVar vid vInfo =
     let sVar =
       SmtEncoder.getEncodingVariableAtCurrentDepth encInfo vInfo in
     let vval =
       SmtEncoder.extractVariableValue encInfo.SmtEncoder.ctx sModel sVar
         vInfo.CompAst.vType
     in Inthash.add vHash vid (convertConstant vval) in
   let hClock cID cInfo =
     let sVar =
       SmtEncoder.getEncodingVariableAtCurrentDepth encInfo cInfo in
     let sFrozenVar =
       SmtEncoder.getClockFrozenVariableAtCurrentDepth encInfo cInfo in
     let vval =
       SmtEncoder.extractVariableValue encInfo.SmtEncoder.ctx sModel sVar
         CompAst.RealType in
     let vFrozenVal =
       SmtEncoder.extractVariableValue encInfo.SmtEncoder.ctx sModel
         sFrozenVar CompAst.BoolType
     in
     match (vval, vFrozenVal) with
       | (CompAst.NumConst f, CompAst.BoolConst b) ->
         Inthash.add cHash cID (b, f)
       | _ -> assert false
   in
       (* type mismatch in these varaibles? *)
   (Inthash.iter hVar gInfo.CompAst.gVars;
    Inthash.iter hClock gInfo.CompAst.gClocks;
    SI.delModel sModel;
    SI.delContext encInfo.SmtEncoder.ctx;
    (vHash, cHash)))
    
let getInitialState gInfo = (*-- 1. Get values for the parameters. --*)
  (*-- 2. Get values for the variables and clocks that satisfy the initial conditions of each of the components. --*)
  let pVals = getParamValues gInfo in
  let (varVals, clockVals) = getVariableAndClockValues pVals gInfo
  in { paramValues = pVals; varValues = varVals; clockValues = clockVals; }
  
(*--- CORE simulator code ---*)
let createEncodingInfoFromState gInfo curState = 
  let ctx = SI.createContext () in 
  let eInfo = {
    SmtEncoder.ctx = ctx ;
    SmtEncoder.pHash = Inthash.create 10;
    SmtEncoder.vHash = IntPairhash.create 50;  (* A hashtable from the variables at particular depth to the corr. variables *)
    SmtEncoder.vClockFrozenHash = IntPairhash.create 50; (* Variable to track whether a clock is frozen or not *)
    SmtEncoder.vTrsTakenHash= IntPairhash.create 10; (* transition ID, depth -> encoding *)
    curDepth = 1;
  }
  in 
  Inthash.iter  (fun _ vi -> ignore(SmtEncoder.makeVar eInfo vi)) gInfo.CompAst.gNondets;
  let valToExpr cVal = 
    match cVal with 
      | FloatVal f -> SI.floatConstant ctx f
      | EnumVal i -> SI.intConstant ctx i
      | BoolVal b -> SI.boolConstant ctx b
  in 
  let hParams pID pVal = 
    Inthash.add eInfo.SmtEncoder.pHash pID (valToExpr pVal)
  in 
  let hVars vID cVal = 
    IntPairhash.add eInfo.SmtEncoder.vHash (vID,1) (valToExpr cVal)
  in
  let hClock cID (b,f) =
    let fExpr = SI.floatConstant  ctx f in 
    let bExpr = SI.boolConstant ctx b in 
    IntPairhash.add eInfo.SmtEncoder.vHash (cID,1) fExpr;
    IntPairhash.add eInfo.SmtEncoder.vClockFrozenHash (cID,1) bExpr
  in
  Inthash.iter hParams curState.paramValues;
  Inthash.iter hVars curState.varValues;
  Inthash.iter hClock curState.clockValues;
  eInfo

let extractRelevantNondetValuesForTrs gInfo eInfo sModel gExpr = 
  let ndHash = Inthash.create 20 in 
  let ndVars = CompAst.collectVariablesFromExpr (fun vi -> vi.CompAst.isNondet) gExpr in 
  let hVarID vid =  (* find value for vid *)
    let vi = CompAst.getNondetVarInfoFromID gInfo vid in 
    let sVar = SmtEncoder.getEncodingVariableAtDepth eInfo vi eInfo.SmtEncoder.curDepth in 
    let vValue = SmtEncoder.extractVariableValue eInfo.SmtEncoder.ctx sModel sVar vi.CompAst.vType in 
    Inthash.replace ndHash vid (convertConstant vValue)
  in 
  Intset.iter hVarID ndVars;
  ndHash
    

    
let chooseEnabledTransition tList gInfo eInfo curState comp  = 
  let ctx = eInfo.SmtEncoder.ctx in 
	(* For each transition, assert the guard and see if the result is satisfiable, *)
	(* if so, record the values for all the nondets and the enabled transition *)
  let handleTrs enabledList ctrs =
    SI.push ctx ;
    let yExpr = SmtEncoder.encodeExpression eInfo ctrs.CompAst.tGuard in 
    SI.assertExpr ctx yExpr;
    match (SI.check ctx) with 
      | SI.SAT -> (
					 (* 1. Extract values for all the nondets *)
	let sModel = SI.getModel ctx in 
	let nHash = extractRelevantNondetValuesForTrs gInfo eInfo sModel ctrs.CompAst.tGuard in 
	SI.delModel sModel;
	SI.pop ctx;
	enabledList@[(nHash,ctrs)]
      )
      | SI.UNSAT
      | SI.FAIL -> ( SI.pop ctx; 
		     enabledList
      )
  in 
  let eList = List.fold_left handleTrs [] comp.CompAst.cTransitions in 
  if (List.length eList) <= 0 
  then ( 
    tList 
  )
  else (
    let (nHash,cTrs) = MyUtils.chooseRandomValFromList eList in 
    (nHash,cTrs)::tList
  )

let updateClockValue nState vid cVal = 
  match cVal with 
    | FloatVal f' -> (
      let (b,f) = try Inthash.find nState.clockValues vid with | Not_found -> assert false in 
      Inthash.replace nState.clockValues vid (b,f')
    )
    | _ -> assert false

let updateVariableValue nState vid cVal = 
  Inthash.replace nState.varValues vid cVal

 let clockIncrement newState cid (b,f) = 
	 if (b) then (
				 Inthash.replace newState.clockValues cid (b, (f+.1.0));
				 
				)


let updateStateForTransition gInfo eInfo curState newState (ndHash, cTrs) = 
  let hUpdate vid uExpr = 
    let cVal = evalExpr (curState,ndHash) uExpr in 
    let vi = CompAst.getVarInfoFromID gInfo vid in 
    if (vi.CompAst.isClock) 
    then (
      updateClockValue newState vi.CompAst.vID	cVal 
    )
    else (
      updateVariableValue newState vi.CompAst.vID cVal	
    )
  in
    
	
  let hClockFreeze vi = 
		try 
			let (_,f) = Inthash.find newState.clockValues vi.CompAst.vID 
			in 
		 		Inthash.replace newState.clockValues vi.CompAst.vID (true,f)
		with | Not_found -> assert false
  in 
  let hClockUnfreeze vi = 
		try let (_,f) = Inthash.find newState.clockValues vi.CompAst.vID 
		in 
		 Inthash.replace newState.clockValues vi.CompAst.vID (false,f)
		with | Not_found -> assert false
  in 
  Inthash.iter hUpdate cTrs.CompAst.tUpdates ;
  List.iter hClockFreeze cTrs.CompAst.tClockFreeze;
  List.iter hClockUnfreeze cTrs.CompAst.tClockUnfreeze
    
    
    
let simulateOneStep gInfo curState = 
	(* 0. Create an encoding info from current state remembering to add fresh variables for non-dets *)
	(* 1. For each of the components, find a list of enabled transitions.*)
	(* 2. Choose one enabled transition from each component. *)
	(* 3. Run the transitions in parallel and obtain the new values of *)
	(*   variables and clocks after the transition *)	
  let eInfo = createEncodingInfoFromState gInfo curState in 
  let hComp tList comp = chooseEnabledTransition tList gInfo eInfo curState comp in 
  let trsListForComps = List.fold_left hComp []  gInfo.CompAst.gComponents in 
  let newState = { 
    paramValues = curState.paramValues;
    varValues = Inthash.copy curState.varValues;
    clockValues = Inthash.copy curState.clockValues
  } in 
	Inthash.iter (clockIncrement newState) newState.clockValues;
  List.iter (updateStateForTransition gInfo eInfo curState newState) trsListForComps;
  SI.delContext eInfo.SmtEncoder.ctx;
  newState
    
let simulate gInfo nSteps =
  let nState = ref (getInitialState gInfo)
  in printState stdout gInfo !nState ;
  Printf.printf "\n";
  for i = 1 to nSteps
  do
    nState := simulateOneStep gInfo !nState;
    printState stdout gInfo !nState;
    print_endline "" 
  done
    
let simAndGenerateTrace gInfo nSteps = 
  let iState = getInitialState gInfo in 
  let stList = ref ([iState]) in 
  let curState = ref (iState) in 
  for i = 0 to (nSteps-2) 
  do
    curState := (simulateOneStep gInfo !curState);
    stList := !curState :: (!stList)
  done;
  Array.of_list ( List.rev !stList)

let computeObjectiveValue gInfo mExpr stArray maxVals minVals = 
  let nSteps = Array.length stArray  in 
  for i = 0 to (nSteps -1 )
  do
    let st = stArray.(i) in
    let ndDummyHash = Inthash.create 1 in 
    let eVal = evalExpr (st,ndDummyHash) mExpr in 
    match eVal with 
      | FloatVal f -> (
	if (maxVals.(i) <= f)
	then maxVals.(i) <- f;
	if  (minVals.(i) > f)
	then minVals.(i) <- f 
      )
      | _ -> (
	Printf.fprintf stderr "Warning --> cannot be asked to optimize a non-real expression. \n"
      )
  done

    
