(*---
  CompAst: This file contains the AST representation for the various components in the system
  
  Author: Sriram Sankaranarayanan < srirams@colorado.edu > 
  --*)

open MyUtils

exception TypeError
exception SemanticError of string

let varCounter = ref 1
let enumCount = ref 1
let compCounter = ref 1
let paramCounter = ref 1
let trsCounter = ref 1
let nProps = ref 0

type globInfo = {
  mutable  gParams: paramInfo Inthash.t; (* All parameters *)
  mutable  gVars: varInfo Inthash.t; (* All globals *)
  mutable  gNondets: varInfo Inthash.t;
  mutable  gClocks: varInfo Inthash.t; (* All clocks *)
  mutable  gEnums: string Inthash.t;  (*All enum symbols *)
  mutable  gAssumes: expr list;
  mutable  gProps: props list;
  mutable  gComponents: compInfo list; (* All the automata components *)  
  mutable  gSyms: symInfo Strhash.t;
  mutable  gTimeStep: float;
  mutable  gMax: maxObjType list;
} 
and maxObjType = {
  mExpr: expr;
  mLo: float;
  mHi: float;
  mStep: float;
  mutable mCur: float;
  mutable limExceeded: bool;
  mCond: expr;
}
and props = {
  prpID: int;
  pExpr: expr;
  mutable pViolated: bool
}
and varInfo = {
  vName: string;
  mutable  vID: int;
  mutable  vCompID: int;
  mutable  vType: typeInfo;
  mutable vRange: rangeInfo;
  mutable isNondet: bool;
  mutable isClock: bool;
  mutable alwaysRunningClock: bool; (* As an optimization *)
}
and typeInfo = 
  | RealType
  | StringType
  | BoolType
  | EnumType 
and expr = 
  | ABoolOp of boolOp * expr * expr
  | ANot of expr
  | ARelOp of relOp * expr * expr 
  | AArithOp of binOp * expr * expr
  | ANeg of expr
  | AVar of varInfo
  | AConst of constInfo
  | ADefined of paramInfo
and boolOp = AAnd | AOr | AImpl
and relOp = AGeq | AGt | AEq | ANeq
and binOp = APlus | AMinus | AStar | ADiv
and paramInfo = {
  pName: string;
  mutable pType: typeInfo;
  mutable  pID: int;
  mutable  pRange: rangeInfo;
  mutable  pIsUserInp: bool;
}

and compInfo = {
  cName: string;
  mutable cID: int;
  mutable cVars: varInfo Inthash.t;
  mutable cInit: expr;
  mutable cTransitions: transition list;
} 

and transition = {
  tID: int;
  mutable tGuard: expr;
  mutable tUpdates: expr Inthash.t;
  mutable tClockFreeze: varInfo list;
  mutable tClockUnfreeze: varInfo list
}

and rangeInfo =  | RangeSpecNone 
		 | RangeSpecInterval of float * float
		 | RangeSpecEnum of int list 
		 | RangeSpecEq of constInfo

and constInfo = | BoolConst of bool
		| StringConst of string
		| NumConst of float
		| EnumConst of int
		| ParamConst of paramInfo


and symInfo = 
  | DeclVar of varInfo 
  | DeclParam of paramInfo
  | DeclEnum of int
	
(*-- Temporary Type for range specifications --*)


let gInfo = {
   gParams = Inthash.create 100;
   gVars = Inthash.create 100;
   gNondets = Inthash.create 100;
   gClocks = Inthash.create 100;
   gSyms = Strhash.create 100;
   gComponents = [];
   gEnums=Inthash.create 100;
   gProps = [];
   gAssumes = [];
   gTimeStep = 1.0;
   gMax = [];
 }

let getParamInfoFromID gInfo pId = 
	try Inthash.find gInfo.gParams pId with Not_found -> raise (Failure "no such parameter found for given id")

let resolveIdentifier str = 
  Strhash.find gInfo.gSyms str
  (*  (\*-- Check if str is a parameter --*\) *)
  (* if (Strhash.mem gInfo.gParams str) *)
  (* then  *)
  (*   begin  *)
  (*     try *)
  (* 	DeclParam(Strhash.find gInfo.gParams str) *)
  (*     with  | Not_found -> assert false *)
  (*   end *)
  (* else *)
  (*   begin *)
  (*     if (Strhash.mem gInfo.gVars str) *)
  (*     then begin *)
  (*      (\*-- Check if it is a variable --*\) *)
  (* 	try *)
  (* 	  DeclVar(Strhash.find gInfo.gVars str) *)
  (* 	with | Not_found -> assert false *)
  (*     end *)
  (*     else  *)
  (* 	 (\*-- Check if it is an enum --*\) *)
  (* 	  DeclEnum (str, (List.find (isEnumMem str) gInfo.gEnums)) *)
  (*   end *)


let eqType tA tB = (tA = tB)

let typeCheckAssign vi (e,eType) =  eqType vi.vType eType
  

let addComponent compName compVars compInitExpr compTransitions = 
  let c = { 
    cName = compName;
    cID = !compCounter;
    cVars = compVars;
    cInit = compInitExpr;
    cTransitions = compTransitions
  } in 
  (* fill in the compID fields *)
  Inthash.iter (
    fun _ vi -> 
      begin 
	vi.vCompID <- c.cID
      end 
  ) c.cVars;
  incr compCounter;
  gInfo.gComponents <- c::gInfo.gComponents

let isBoolType cType = 
  cType = BoolType

let declareNewEnumSymbols strList = 
  (*-- Insert all enum strings as symbols --*)
  List.map (
    fun str ->   (
      ( try 
	  begin
	    match (Strhash.find gInfo.gSyms str) 
	    with 
	      | DeclEnum _ -> ()
	      | DeclVar vi -> raise (SemanticError ("Enum symbol already defined as a variable: "^str))  
	      | DeclParam pi -> raise (SemanticError ("Enum symbol already defined as a parameter: "^str))  	    
	  end 
	with | Not_found -> ()
      );
      let x = !enumCount in 
      incr enumCount;
      Strhash.replace gInfo.gSyms str (DeclEnum x);
      Inthash.replace gInfo.gEnums x str;
      x
    ) 
  ) strList
    
    

(* let resolveEnumType strList =  *)
(*   let checkAllEnumsAndFindMatch lst =  *)
(*     let eIRef = match (List.hd lst)  with  *)
(*       | None -> None *)
(*       | Some (DeclEnum(str,eI)) -> Some eI *)
(*       | Some (DeclVar vi) -> raise  *)
(* 	(SemanticError  *)
(* 	   ("Enumlist contains strings already declared as variable:"^(vi.vName)) *)
(* 	) *)
(*       | Some (DeclParam pi) -> raise  *)
(* 	( *)
(* 	  SemanticError  *)
(* 	    ("Enumlist contains strings already declared as parameter:"^(pi.pName)) *)
(* 	) *)
	
(*     in  *)
(*     List.iter ( *)
(*       fun d -> begin  *)
(* 	match d with  *)
(* 	  | None ->  *)
(* 	    if not ( eIRef = None)   *)
(* 	    then (raise (SemanticError "Enum list contains a mix of unknown strings and previously declared enums")) *)
(* 	  | Some (DeclEnum (str,eI)) ->  *)
(* 	    ( match eIRef with  *)
(* 	      | None ->  raise  *)
(* 		( *)
(* 		  SemanticError  *)
(* 		    ("Enum list contains a mix of unknown strings and previously declared enums. Mismatch example: "^str)  *)
(* 		) *)
(* 	      | Some (eI') ->  *)
(* 		if not (eI.eID = eI'.eID) *)
(* 		then raise  *)
(* 		  ( *)
(* 		    SemanticError *)
(* 		      ("Enum list contains elements from two enums. Mismatch example: "^str ) *)
(* 		  ) *)
(* 	    ) *)
(* 	  | Some (DeclVar vi) -> raise ( SemanticError ("Enum element: "^(vi.vName)^" already declared as a var. ")) *)
(* 	  | Some(DeclParam pi) -> raise ( SemanticError ("Enum element: "^(pi.pName)^" already declared as a param ")) *)
(* 	  | _ -> assert false *)
(*       end *)
(*     ) (List.tl lst); *)
(*     eIRef *)
(*   in  *)
(*   (\*--  *)
(*     1. Check if any of the string in the list intersects with known enums. *)
(*     --*\) *)
(*   match strList with *)
(*     | [] -> raise (SemanticError "Enum type declared with no members") *)
(*     | hd::tail ->  *)
(*       let rList = List.map ( *)
(* 	fun s -> begin  *)
(* 	  try Some (Strhash.find gInfo.gSyms s) with  *)
(* 	    | Not_found -> None  *)
(* 	end *)
(*       ) strList in  *)
(*       (match rList with  *)
(* 	| [] -> assert false *)
(* 	| hd::tail -> ( *)
(* 	  let eI = checkAllEnumsAndFindMatch rList in  *)
(* 	  match eI with  *)
(* 	    | None -> declareNewEnum strList *)
(* 	    | Some eI -> eI *)
	  
(* 	) *)
(*       ) *)
      
  
let declareVariable str typ rSpec = 
    let vi = {
      vName = str;
      vID = !varCounter;
      vCompID = -1; (*-- to be filled in later --*)
      vType = typ;
      vRange = rSpec;
      isNondet = false;
      isClock = false;
      alwaysRunningClock = false;
    }
    in 
    incr varCounter;
    if (Strhash.mem gInfo.gSyms str)
    then 
      raise (SemanticError ("Redeclaration of variable "^str));
    Inthash.replace gInfo.gVars vi.vID vi;
    Strhash.replace gInfo.gSyms str (DeclVar vi);
    match (typ, rSpec)  with
      | _ , RangeSpecNone -> vi
      | EnumType, (RangeSpecEnum sList) -> (
	vi.vType <- (EnumType);
	vi.vRange <- rSpec;
	vi
      )
      | RealType, (RangeSpecInterval (l,u)) -> 
	( vi.vRange<- rSpec;
	  vi
	) 
      | BoolType,  (RangeSpecEq (BoolConst _))
      | EnumType, (RangeSpecEq (EnumConst _)) 
      | RealType, (RangeSpecEq (NumConst _)) -> 
	(
	  vi.vRange <- rSpec;
	  vi
	)

      |  _ ->  (raise (SemanticError ( ("Range Specification Type Mismatch: ")^vi.vName)))
      

let declareParam pN pIsUserDefined pT  prSpec = 
  try 
    let _ = resolveIdentifier pN in 
    raise (SemanticError ("Parameter "^pN^" already declared."))
  with 
    | Not_found -> ();
      let pI = { 
	pName = pN;
	pType  = pT;
	pID = !paramCounter;
	pRange = prSpec;
	pIsUserInp = pIsUserDefined;
      } in 
	incr paramCounter;
	Strhash.replace gInfo.gSyms pN (DeclParam (pI));
	Inthash.replace gInfo.gParams pI.pID pI
	

let declareClock clkName clkRange =
  let vi = {
       vName = clkName;
       vID = !varCounter;
       vCompID = -1; (*-- to be filled in later --*)
       vType = RealType;
       vRange = clkRange;
       isNondet = false;
       isClock = true;
       alwaysRunningClock = true;
    }
    in 
     incr varCounter;
     if (Strhash.mem gInfo.gSyms clkName)
     then raise (SemanticError ("Redeclaration of clock variable "^clkName));
     Inthash.replace gInfo.gClocks vi.vID vi;
     Strhash.replace gInfo.gSyms clkName (DeclVar vi);
     vi

let declareNondet name nType nRange =
     let vi = {
       vName = name;
       vID = !varCounter;
       vCompID = -1; (*-- to be filled in later --*)
       vType = nType;
       vRange = nRange;
       isNondet = true;
       isClock = false;
       alwaysRunningClock = false;
    }
    in 
     incr varCounter;
     if (Strhash.mem gInfo.gSyms name)
     then 
       raise (SemanticError ("Redeclaration of nondet variable "^name));
     Inthash.replace gInfo.gNondets vi.vID vi;
     Strhash.replace gInfo.gSyms name (DeclVar vi);
     vi

let makeTrs e h fList ufList =
  let trsCount = !trsCounter in 
  incr trsCounter;
  let hFrozenClock vi = 
    assert (vi.isClock);
    vi.alwaysRunningClock <- false
  in 
  List.iter hFrozenClock fList;
  {
    tID = trsCount;
    tGuard = e;
    tUpdates = h;
    tClockFreeze = fList;
    tClockUnfreeze = ufList
  }
let addAssumeSpec e = 
  gInfo.gAssumes <- e::gInfo.gAssumes

let addAssertSpec e =
  let prp = { prpID = !nProps; pExpr = e; pViolated = false; } in 
  incr nProps;
  gInfo.gProps <- prp::gInfo.gProps

let checkRangeTypeMatch str rType rSpec = ()

let string_of_boolOp b =  match b with | AAnd -> "and" | AOr -> "or" | AImpl -> "impl"

let string_of_relOp r = (match r with | AGeq -> "GEQ" | AGt -> "GT" | AEq -> "EQ" | ANeq -> "NEQ" )

let string_of_arithOp b = (match b with | APlus -> "+" | AMinus -> "-" | AStar -> "*" | ADiv -> "/" )

let string_of_constinfo ci = match ci with 
	      | BoolConst b -> if (b) then "bool:true" else "bool:false" 
	      | StringConst str -> "string:"^str
	      | NumConst f -> "num:"^(string_of_float f)
	      | EnumConst i -> "enum:"^(string_of_int i)^"-->"^(Inthash.find gInfo.gEnums i)
	      | ParamConst pi -> "param:"^(pi.pName)

let string_of_enum_id gInfo i = 
  try 
    Inthash.find gInfo.gEnums i 
  with | Not_found -> "Unknown enum: something shady just happened"
    
let paramConst pInfo = 
  AConst (ParamConst pInfo)

let enumConst s = 
  AConst (EnumConst s)

let numConst f = 
  AConst (NumConst f)


let exprOfRange ex0 r = 
  match r with 
    | RangeSpecNone -> AConst (BoolConst true)
    | RangeSpecInterval (u, v) -> 
      let lbExpr =  
	ARelOp (AGeq, ex0, (numConst u)) 
      in 
      let ubExpr = 
	ARelOp (AGeq, (numConst v),ex0)
      in 
      (ABoolOp (AAnd, (lbExpr), ( ubExpr )  ))
    | RangeSpecEq cInfo -> 
      ARelOp (AEq, ex0,	(AConst cInfo))
     
    | RangeSpecEnum sList ->   (* Or of all the options *)
	List.fold_left (fun e s -> 
			ABoolOp (AOr, e, (ARelOp  (AEq, ex0, (enumConst s) ) ) )
		     ) (AConst (BoolConst false)) sList




let getDefinedExpression pI = exprOfRange (paramConst pI) pI.pRange

let getDomainExpressionForVar vi = exprOfRange (AVar vi) vi.vRange


let getVariableNameFromID gInfo vid = 
  let vi = if (Inthash.mem gInfo.gVars vid) 
  then Inthash.find gInfo.gVars vid 
  else (
    if (Inthash.mem gInfo.gNondets vid)
    then Inthash.find gInfo.gNondets vid
    else 
      if (Inthash.mem gInfo.gClocks vid)
      then Inthash.find gInfo.gClocks vid
      else assert false
  ) in 
    vi.vName


let transitionFreezesClock trs vi = 
  List.exists (fun vi' -> vi.vID = vi'.vID) trs.tClockFreeze


let transitionUnfreezesClock trs vi = 
  List.exists (fun vi' -> vi.vID = vi'.vID) trs.tClockUnfreeze

let addMaxObj e l u s c =
  let m = { mExpr = e;
	    mLo= l; 
	    mHi = u; 
	    mStep = s;
	    mCur = l -. 1.0; 
	    mCond = c;
	    limExceeded = false
	  }
  in 
    gInfo.gMax <- m::gInfo.gMax


let printMaxValue gM = 
  if (gM.limExceeded) 
  then  Printf.printf "\nMax value out of hi range. \n"
  else if (gM.mCur >= gM.mLo) 
  then Printf.printf "\nMAX VALUE FOUND = %f \n" gM.mCur
  else 
    Printf.printf "\nSearch did not find any value in range. \n"
    
      
let printMaxValues () =
  List.iter printMaxValue gInfo.gMax

exception FoundParam of paramInfo


let getParamInfoFromName paramName = 
  let hParam _ pInfo = 
    if (String.compare pInfo.pName paramName) = 0
    then raise (FoundParam pInfo)
  in 
  try 
    Inthash.iter hParam gInfo.gParams ;
    raise (Failure ("Unknown parameter name :"^paramName))
  with 
    | FoundParam pInfo -> pInfo

let getFloatConstFromParamString pStr = 
  NumConst (float_of_string pStr)

exception FoundEnumID of int

let getEnumConstFromParamString pStr = 
  let hEnum i eStr = 
    if (String.compare pStr eStr) = 0 
    then raise (FoundEnumID i)
  in 
  try 
    Inthash.iter hEnum gInfo.gEnums ;
    raise (Failure ("Cannot find enum constant "^pStr) )
  with 
    | FoundEnumID i -> (EnumConst i)

let getBoolConstFromParamString pStr = 
  if (String.compare pStr "true") = 0
  then (BoolConst true)
  else 
    if (String.compare pStr "false") = 0
    then 
      BoolConst false
    else 
      raise (Failure ("Unable to parse bool constant: "^pStr))


let setParamValue paramName pString =
  if ( (String.length paramName) <= 0)  || ((String.length pString) <= 0)
  then ()
  else (
    Printf.printf " -> Setting parameter %s to %s \n" paramName pString;
    let pI = getParamInfoFromName paramName  in 
    match (pI.pType) with 
      | RealType -> pI.pRange <- RangeSpecEq (getFloatConstFromParamString pString)
      | EnumType -> pI.pRange <- RangeSpecEq (getEnumConstFromParamString pString)
      | BoolType -> pI.pRange <- RangeSpecEq (getBoolConstFromParamString pString) 
      | _ -> raise (Failure "Type mismatch between parameter name and value it is set to. ")
  )


let readParametersFromFile oc = 
  let rec rFile () = 
    try 
      Scanf.sscanf (input_line oc) "%s %s " (fun a b -> setParamValue a b);
      rFile ()
    with | End_of_file  -> ()
  in 
  rFile ()


let chooseInitialParamValueAtRandom pI = (*- Choose a random value from the parameter range -*)
  match pI.pRange with 
		| RangeSpecNone  -> raise Not_found
		| RangeSpecInterval(l,u) -> NumConst (chooseRandomValFromRange l u)
		| RangeSpecEnum iList -> EnumConst (chooseRandomValFromList iList)
		| RangeSpecEq  cInfo -> cInfo

let rec collectVariablesFromExpr filterFunc e = match e with 
	| ABoolOp(_,e1,e2)
	| ARelOp (_,e1,e2)
	| AArithOp(_,e1,e2) -> Intset.union (collectVariablesFromExpr filterFunc e1) (collectVariablesFromExpr  filterFunc e2)
  | ANot e
  | ANeg e -> collectVariablesFromExpr filterFunc e
  | AVar vi -> if (filterFunc vi) then Intset.singleton (vi.vID) else Intset.empty
  | _ -> Intset.empty

let getVarInfoFromID gInfo vid = 
   if (Inthash.mem gInfo.gVars vid) 
   then 		 Inthash.find gInfo.gVars vid 
	 else ( 
		if (Inthash.mem gInfo.gClocks vid) 
	  then Inthash.find gInfo.gClocks vid 
		else ( 
				if (Inthash.mem gInfo.gNondets vid)
			  then Inthash.find gInfo.gNondets vid 
				else raise (Failure ("Cannot find variable "^(string_of_int vid)))
				)
		)
	

let getNondetVarInfoFromID gInfo vid = 
	try Inthash.find gInfo.gNondets vid with  Not_found -> raise (Failure "Unknown variable id" )
	
