#ifndef D__PRESCRIPTION__DATA__HH__
#define D__PRESCRIPTION__DATA__HH__

/*---
  Class: PrescriptionData
  Author: Sriram Sankaranarayanan <srirams@colorado.edu>
  --*/
#include "myUtils.hh"

namespace IpSim {

  typedef enum { CONTINUOUS, BOLUS, CONT_BOLUS, PCA, UNDEF } delivery_mode_t;
  /*---
    CONTINUOUS MODE: deliver a continuous dosage. 
    parameters:  Rate to be delivered,  time.
    
    BOLUS MODE: deliver a bolus dosage.
    parameters: total volume to be delivered.

    CONT_BOLUS_MODE: deliver an initial bolus and then deliver a continuous dosage.
    parameters: volume for the bous, continuous rate,  time for continuous.
    
    PCA: deliver  a basal continuous rate and a bolus upon request.
    parameters: volume for the bous, continuous rate,  time for continuous.
    total # of bolus dosages allowed.
    minimum time to elapse between boli.
    ---*/
  
  class PrescriptionData {
  public:
    /*-- Basic prescription data for the patient used to program the pump. --*/
    delivery_mode_t dMode; /*- What delivery mode? -*/
    
    /*-- Data about drug loaded in syringe --*/
    string drugName; /*-- default to fentanyl --*/
    double volumeInSyringe; /*-- Valid range: [1,100] milliliters --*/
    double concentration; /*-- Valid range: [0.1,100] mg/ml --*/

    /*-- Data for each delivery mode --*/
    double continuousRate; /*-- 
			     Valid range: [0.01,10]
			     basal rate of delivery (valid: CONT, CONT_BOLUS, PCA modes only). 
			     Units: ml/min 
			     --*/
    double totTime; /*-- 
		      Valid range: [1,10000]
		      total time to deliver dosage (min). 
		      valid: CONT, CONT_BOLUS, PCA
		   --*/
    double volBolus; /*-- 
		       Valid range: [0.1,100]
		       How much volume to deliver for a bolus.
		       valid: PCA, CONT_BOLUS only
		       Units: ml 
		       --*/
    int numPCABoli; /*--
		      Valid range: [1,100]
		      Number of PCA bolus allowed (max) --*/
    double timeBetweenBoli;  /*--
			       Valid range: [0.1,1000]
			       minimum time allowed between two successful bolus dosages. 
			       --*/

    PrescriptionData():dMode(UNDEF), 
		       drugName("unknown"),
		       volumeInSyringe(-1.0),
		       concentration(-1.0),
		       continuousRate(-1.0),
		       totTime(-1.0),
		       volBolus(-1.0),
		       numPCABoli(-1),
		       timeBetweenBoli(-1.0){}

    bool dOK(){
      /*-- Is everything necessary been set? ---*/
      if ( OUT_OF_RANGE(volumeInSyringe,1.0,100.0) ) || OUT_OF_RANGE(concentration,0.1,100.0)){
	std::cout << " Better set syringe volume and concentration! Cannot accept prescription. "<< std::endl;
	return false;
      }
      switch (dMode){
      case UNDEF:
	{ 
	  std::cout << "Need to set a mode. "<< std::endl;  
	  return false; 
	}  
	break;
      case CONTINUOUS: 
	{
	  if (OUT_OF_RANGE(continuousRate,0.01,10.0) || OUT_OF_RANGE(totTime,1,10000)){
	    std::cout << " Continuous mode needs to set rate/time in a valid range. " << std::endl;
	    return false;
	  }
	  break;
	}
      case BOLUS:
	{
	  if (OUT_OF_RANGE(volBolus,0.1,100.0)){
	    std::cout << "Bolus mode needs to specify a valid volume to be infused "<< std::endl;
	    return false;
	  }
	  if (volBolus > volumeInSyringe){
	    std::cout << "Bolus volume exceeds declared volume in syringe " << std::endl;
	    return false;
	  }
	}
	break;
      case CONT_BOLUS_MODE: 
	{
	  if (OUT_OF_RANGE(continuousRate,0.01,10.0) || OUT_OF_RANGE(totTime,1,10000)){
	    std::cout << " Continuous+bolus mode needs to set rate as well as time. " << std::endl;
	    return false;
	  }

	  if (OUT_OF_RANGE(volBolus,0.1,100.0)){
	    std::cout << "Continuous + Bolus mode needs to specify a volume to be infused "<< std::endl;
	    return false;
	  }
	  if (volBolus > volumeInSyringe){
	    std::cout << "Continuous + Bolus volume exceeds declared volume in syringe " << std::endl;
	    return false;
	  }
	  
	}
	break;

      case PCA: 
	 
	if (OUT_OF_RANGE(continuousRate,0.01,10.0) || OUT_OF_RANGE(totTime,1,10000)){
	  std::cout << " PCA mode needs to set rate as well as time. " << std::endl;
	  return false;
	}

	if (OUT_OF_RANGE(volBolus,0.1,100.0)){
	  std::cout << "PCA mode needs to specify a bolus volume to be infused "<< std::endl;
	  return false;
	}
	if (volBolus > volumeInSyringe){
	  std::cout << "PCA mode  bolus volume exceeds declared volume in syringe " << std::endl;
	  return false;
	}
	if (OUT_OF_RANGE(numPCABoli,1,100)){
	  std::cout << "PCA mode  num bolus needs to be set. " << std::endl;
	  return false;
	}
	if (OUT_OF_RANGE(timeBetweenBoli, 0.1,1000.0)){
	   std::cout << "PCA mode time between bolus requests is too small. " << std::endl;
	  return false;
	} 
      }
      return true;
    }    
  };
};

#endif
