(*---
  Encode executions in SMT. We will use  a generic SMT encoding interface that can be switched between Z3 and Yices
  --*)

open MyUtils 
module C = CompAst
module S = SmtZ3Interface

let debug = ref false

    

type encodingInfo = {
  ctx: S.context;
  pHash: S.expr Inthash.t; (* A hashtable from the parameters to the variables *)
  vHash: S.expr IntPairhash.t;  (* A hashtable from the variables at particular depth to the corr. variables *)
  vClockFrozenHash: S.expr IntPairhash.t; (* Variable to track whether a clock is frozen or not *)
  vTrsTakenHash: S.expr IntPairhash.t; (* transition ID, depth -> encoding *)
  mutable curDepth: int
}


let createEncodingInfo () = {
  ctx = (S.createContext () );
  pHash = Inthash.create 100;
  vHash = IntPairhash.create 100;
  vClockFrozenHash = IntPairhash.create 100;
  vTrsTakenHash = IntPairhash.create 100;
  curDepth = 0;
}

let getClockFrozenVariableAtDepth eInfo vi depth = 
  assert(vi.C.isClock);
  try 
    IntPairhash.find eInfo.vClockFrozenHash (vi.C.vID, depth)
  with 
    | Not_found -> (
      Printf.fprintf stderr "Encoding Error: Did not encode clock frozen variable for clock %s at depth %d \n" vi.C.vName depth;
      assert false
    )



let getClockFrozenVariableAtCurrentDepth eInfo vi = 
  getClockFrozenVariableAtDepth eInfo vi eInfo.curDepth

let getClockFrozenVariableAtPreviousDepth eInfo vi = 
    getClockFrozenVariableAtDepth eInfo vi (eInfo.curDepth-1)

let getEncodingVariableAtDepth eInfo vi depth = 
  try 
    IntPairhash.find eInfo.vHash (vi.C.vID, depth)
  with 
    | Not_found -> (
	Printf.fprintf stderr "Encoding Error: Did not encode variable %s at depth %d \n" vi.C.vName depth;
	assert false
      )

let getEncodingVariableAtCurrentDepth eInfo vi = 
  getEncodingVariableAtDepth eInfo vi eInfo.curDepth

let getEncodingVariableAtPreviousDepth eInfo vi = 
  getEncodingVariableAtDepth eInfo vi (eInfo.curDepth-1)

let getEncodingVariableForParameter eInfo pi = 
  try 
    Inthash.find eInfo.pHash pi.C.pID
  with 
    | Not_found -> (
      Printf.fprintf stderr "Could not find encoding variable for param. %s \n" pi.C.pName;
      assert false
    )

  
let rec encodeExpression eInfo expr = 
  match expr with 
    | C.ABoolOp (b, e1, e2) -> 
      begin 
	let sE1 = encodeExpression eInfo e1 in 
	let sE2 = encodeExpression eInfo e2 in 
	match (b) with 
	  | C.AAnd -> S.andExpr eInfo.ctx   sE1 sE2  
	  | C.AOr -> S.orExpr eInfo.ctx   sE1 sE2 
	  | C.AImpl -> S.implExpr eInfo.ctx sE1 sE2
      end
    | C.ANot e -> let sE = encodeExpression eInfo e in  S.negateExpr eInfo.ctx sE
    | C.ARelOp (r,e1,e2) ->  begin 
      let sE1 = encodeExpression eInfo e1 in 
      let sE2 = encodeExpression eInfo e2 in 
      match (r) with 
	| C.AGeq -> S.geqExpr eInfo.ctx sE1 sE2
	| C.AGt -> S.gtExpr eInfo.ctx sE1 sE2
	| C.AEq -> S.eqExpr eInfo.ctx sE1 sE2
	| C.ANeq  -> S.neqExpr eInfo.ctx sE1 sE2
    end
    | C.AArithOp (b,e1,e2) -> 
      begin
	let sE1 = encodeExpression eInfo e1 in 
	let sE2 = encodeExpression eInfo e2 in 
	match (b) with 
	  | C.APlus -> S.plusExpr eInfo.ctx sE1 sE2 
	  | C.AMinus -> S.minusExpr eInfo.ctx  sE1 sE2
	  | C.AStar -> S.starExpr eInfo.ctx sE1 sE2
	  | C.ADiv -> S.divExpr eInfo.ctx sE1 sE2
      end
    | C.ANeg e -> let sE = encodeExpression eInfo e in 
		S.negateExpr eInfo.ctx sE
    | C.AVar vi -> getEncodingVariableAtCurrentDepth eInfo vi
    | C.AConst cI -> begin
      match cI with 
	| C.BoolConst b -> S.boolConstant eInfo.ctx b
	| C.NumConst f -> S.floatConstant eInfo.ctx f
	| C.EnumConst str -> S.intConstant eInfo.ctx str
	| C.ParamConst pI -> 
			       getEncodingVariableForParameter eInfo pI
			
	| C.StringConst str -> assert false
    end
    | C.ADefined pi -> begin
      let e' = C.getDefinedExpression pi in 
      encodeExpression eInfo e'
    end
      

let assertExpr eInfo e = 
  let sE = encodeExpression eInfo e in
    if (!debug) 
    then ( 
      S.ppExpr eInfo.ctx sE
    );
    S.assertExpr eInfo.ctx sE


let makeParamVars eInfo gI = 
  let makeParamVar _ pI =
    let pName = pI.C.pName in 
      ignore (
	let pVar = S.createVariable eInfo.ctx pName pI.C.pType in 
	  Inthash.replace eInfo.pHash pI.C.pID pVar ;
	  let ex = C.getDefinedExpression pI in 
	    assertExpr eInfo ex
      )
  in 
    Inthash.iter makeParamVar gI.C.gParams

let makeVar eInfo vi  = 
  let varName = Printf.sprintf "%s__%d" vi.C.vName eInfo.curDepth in 
  let pVar = S.createVariable eInfo.ctx (varName) vi.C.vType in 
    IntPairhash.replace eInfo.vHash (vi.C.vID, eInfo.curDepth) pVar;
    begin 
      match vi.C.vRange with 
	| C.RangeSpecNone -> ()
	| _ -> let rExpr = C.getDomainExpressionForVar vi in 
	    assertExpr eInfo rExpr
    end;
    pVar

let makeFrozenVar eInfo vi = 
  let pVar = S.createVariable eInfo.ctx (Printf.sprintf "%s__%d_RUNNING" vi.C.vName eInfo.curDepth) C.BoolType  in
    IntPairhash.replace eInfo.vClockFrozenHash  (vi.C.vID, eInfo.curDepth) pVar;
    if (eInfo.curDepth = 0) then 
      begin
	if (!debug) 
	then ( S.ppExpr eInfo.ctx pVar; flush stdout); 
	S.assertExpr eInfo.ctx pVar
      end; (*-- Start with clock running --*)
    pVar


let makeVarsAtDepth gInfo eInfo = 
  Inthash.iter  (fun _ vi ->  ignore(makeVar eInfo vi)) gInfo.C.gVars;
  Inthash.iter  (fun _ vi -> ignore(makeVar eInfo vi)) gInfo.C.gNondets;
  Inthash.iter  (fun _ vi -> ignore(makeVar eInfo vi)) gInfo.C.gClocks;
  Inthash.iter  (fun _ vi -> ignore(makeFrozenVar eInfo vi)) gInfo.C.gClocks



(*--- 

  Encoding initialization:
    1. Create parameters and add all the assertions involving the parameters.
    2. Create variables for length ZERO.
    3. Assert the initial conditions.
  Encoding depth increase:
    Create variables for new depth.
    For each component: 
           create transition variables
           encode that some transition is taken.
           encode the guard condition for the transition.
           encode the updates.
           encode variable preservations.
    
    Assert property violations.
  
---*)


(*-- Create parameters and add all the assertions involving the parameters. --*)
let createAllParameters gInfo eInfo =
  makeParamVars eInfo gInfo;
  List.iter (assertExpr eInfo) gInfo.C.gAssumes


let createVariablesAtCurrentDepth gInfo eInfo = 
  makeVarsAtDepth gInfo eInfo

(*-- Assert the initial conditions. --*)

let assertInitialConditionsForComponent gInfo eInfo cI = 
  assert(eInfo.curDepth = 0);
  assertExpr eInfo  cI.C.cInit

let assertInitialConditions gInfo eInfo = 
  List.iter (assertInitialConditionsForComponent gInfo eInfo) gInfo.C.gComponents
    

(*-- Create transition variables for component --*)

type compEncodingData = {
  mutable eComp: C.compInfo;
  mutable takenVars: (C.transition * S.expr) list; (*-- variables that associate transitions with their taken variables --*)
  mutable defaultTakenVar: S.expr;
  (*-- other information about which variables are updated in which transitions ---*)
}

let createEncodingDataForComponent eInfo comp = 
  let trsCounter = ref 0 in 
  let tVars = 
    List.map ( fun trs -> 
		 let x = !trsCounter in 
		   incr trsCounter;
		   let takenVar = S.createVariable eInfo.ctx 
		     (Printf.sprintf "taken_%d_%d_%d" comp.C.cID x eInfo.curDepth) 
		     C.BoolType  
		   in 
		     IntPairhash.replace eInfo.vTrsTakenHash (trs.C.tID,eInfo.curDepth) takenVar;
		     ( trs, takenVar )
	     ) comp.C.cTransitions 
  in 
  let dVar = 
    S.createVariable eInfo.ctx (Printf.sprintf "taken_%d_def_%d" comp.C.cID eInfo.curDepth) C.BoolType 
  in 
    IntPairhash.replace eInfo.vTrsTakenHash ((-comp.C.cID), eInfo.curDepth) dVar;
    { eComp = comp;
      takenVars = tVars;
      defaultTakenVar = dVar }
      

(* OR all the taken variables *)
let encodeAtLeastOneTransitionTaken eInfo cData = 
  let oExpr = 
    List.fold_left (fun e (trs,tVar) -> S.orExpr eInfo.ctx tVar e) cData.defaultTakenVar cData.takenVars in 
    if (!debug) 
    then ( S.ppExpr eInfo.ctx oExpr; flush stdout); 
    S.assertExpr eInfo.ctx oExpr

(* For every pair of trsTaken Varibles, encode (~v1 \/ ~v2) *)
let encodeAtMostOneTransitionTaken eInfo cData = 
  let hTrsPair v1 v2 = 
    let sE = S.orExpr eInfo.ctx (S.negateExpr eInfo.ctx v1) (S.negateExpr eInfo.ctx v2) in 
      if (!debug) 
      then ( S.ppExpr eInfo.ctx sE; flush stdout); 
      S.assertExpr eInfo.ctx sE
      
  in 
  let hTrs (t1,v1) = 
    List.iter (fun (t2,v2) -> if not (t1.C.tID <= t2.C.tID) then hTrsPair v1 v2) cData.takenVars;
    hTrsPair v1 cData.defaultTakenVar
  in 
  List.iter hTrs cData.takenVars



(* Encode Guard Implication *)

let encodeGuardImplication eInfo cData =
  eInfo.curDepth <- eInfo.curDepth -1;
  let hPair (t1,v1) = 
    let gExpr = encodeExpression eInfo t1.C.tGuard in 
    let sE =  S.implExpr eInfo.ctx v1 gExpr in  
      if (!debug) 
      then ( S.ppExpr eInfo.ctx sE; flush stdout); 
      S.assertExpr eInfo.ctx (
	sE
      )
  in
  let someGuardEnabled = List.fold_left 
    ( fun e (t,v) -> 
	let gExpr = encodeExpression eInfo t.C.tGuard in 
	  S.orExpr eInfo.ctx e gExpr
    ) (S.boolConstant eInfo.ctx false) cData.takenVars 
  in 
  let noGuardEnabled = S.negateExpr eInfo.ctx someGuardEnabled in 
    List.iter hPair cData.takenVars ;
    let sE =  S.implExpr eInfo.ctx cData.defaultTakenVar noGuardEnabled in 
      if (!debug) 
      then ( S.ppExpr eInfo.ctx sE; flush stdout); 
      S.assertExpr eInfo.ctx sE;
      eInfo.curDepth <- eInfo.curDepth +1






let encodeVarPreserved eInfo vi = 
  let sVarPrime = getEncodingVariableAtCurrentDepth eInfo vi in 
  let sVar = getEncodingVariableAtPreviousDepth eInfo vi in 
    S.eqExpr eInfo.ctx sVarPrime sVar

let encodeClockIncrement eInfo gInfo vi = 
  assert(vi.C.isClock);
  let f = gInfo.C.gTimeStep in 
  let fExpr = S.floatConstant eInfo.ctx f in 
  let cVar = getEncodingVariableAtPreviousDepth eInfo vi in 
  let iExpr = S.plusExpr eInfo.ctx cVar fExpr in 
  let cVarPrime = getEncodingVariableAtCurrentDepth eInfo vi in 
  let eqExpr = S.eqExpr eInfo.ctx cVarPrime iExpr in 
  if (vi.C.alwaysRunningClock)
  then eqExpr
  else (
    let sVar = getClockFrozenVariableAtPreviousDepth eInfo vi in 
    let tExpr = S.implExpr eInfo.ctx sVar eqExpr in 
    let elseExpr = S.implExpr eInfo.ctx (S.negateExpr eInfo.ctx sVar) (S.eqExpr eInfo.ctx cVar cVarPrime) in 
    S.andExpr eInfo.ctx tExpr elseExpr
  )
  

(* Encode Updates For Transitions *)

let encodeUpdates eInfo gInfo cVars (trs,vTaken) =
 
  let trsUpdateExpr = ref (S.boolConstant eInfo.ctx true) in 
  let hPres vID vi = 
    if not (Inthash.mem trs.C.tUpdates vID)
    then (
      if (vi.C.isNondet) then ()
      else  (
	let eExpr = if ( vi.C.isClock) 
	  then encodeClockIncrement eInfo gInfo vi
	  else encodeVarPreserved  eInfo vi
	in 
	trsUpdateExpr := S.andExpr eInfo.ctx eExpr !trsUpdateExpr
      )
    )
  in 
  let hClockFreezeVarPreserved vID vi = 
    if (vi.C.isClock) 
    then (
      if (C.transitionFreezesClock trs vi) || (C.transitionUnfreezesClock trs vi)
      then ()
      else (
	let eqExpr = S.eqExpr eInfo.ctx (getClockFrozenVariableAtCurrentDepth eInfo vi) (getClockFrozenVariableAtPreviousDepth eInfo vi) in 
	trsUpdateExpr := S.andExpr eInfo.ctx eqExpr !trsUpdateExpr	
      )
    ) 
  in 
  let hUpdate vID uExpr = 
    eInfo.curDepth <- eInfo.curDepth -1;
    let sExpr = encodeExpression eInfo uExpr in 
      eInfo.curDepth <- eInfo.curDepth +1;
      let vi = 
	try 
	  Inthash.find cVars vID 
	with 
	  | Not_found -> (
	      Printf.fprintf stderr "Variable %s is updated in a different component from its declarations \n" (C.getVariableNameFromID gInfo vID) ; 
	      assert false
	    )
      in 
      let sVar = getEncodingVariableAtCurrentDepth eInfo vi in 
      let eExpr = S.eqExpr eInfo.ctx sVar sExpr in 
	trsUpdateExpr := S.andExpr eInfo.ctx eExpr !trsUpdateExpr
  in 
  let hClockFreeze vi = 
    try 
      let cVar = getClockFrozenVariableAtCurrentDepth eInfo vi in 
      let fExpr = S.negateExpr eInfo.ctx cVar in 
	trsUpdateExpr := S.andExpr eInfo.ctx fExpr !trsUpdateExpr 
    with  | Not_found -> assert false
  in 
  let hClockUnfreeze vi = 
    try 
      let cVar = getClockFrozenVariableAtCurrentDepth eInfo vi in 
	trsUpdateExpr := S.andExpr eInfo.ctx cVar !trsUpdateExpr 
    with  | Not_found -> assert false
  in 

  Inthash.iter hUpdate trs.C.tUpdates;
  List.iter hClockFreeze trs.C.tClockFreeze;
  List.iter hClockUnfreeze trs.C.tClockUnfreeze;
  Inthash.iter hPres cVars;
  Inthash.iter hClockFreezeVarPreserved cVars;
  let sE =  S.implExpr eInfo.ctx vTaken !trsUpdateExpr in 
  if (!debug) 
  then ( S.ppExpr eInfo.ctx sE; flush stdout); 
  S.assertExpr eInfo.ctx ( sE  )
    


(* Encode Update for default transition *)
let encodeUpdateForDefault eInfo gInfo defTakenVar compVars = 
  let trsUpdateExpr = ref (S.boolConstant eInfo.ctx true) in 
  let hVar _ vi = 
    if (vi.C.isNondet) 
    then () (*-- no update for nondets --*)
    else (
      if (vi.C.isClock)  
      then (*-- if clock is running then it is incremented by the time step --*)
	(
	  let sE = encodeClockIncrement eInfo gInfo vi in 
	  trsUpdateExpr := S.andExpr eInfo.ctx sE !trsUpdateExpr;
	  (*-- encode the update of clock running variables --*)
	  let eqExpr = S.eqExpr eInfo.ctx (getClockFrozenVariableAtCurrentDepth eInfo vi) 
	    (getClockFrozenVariableAtPreviousDepth eInfo vi) in 
	  trsUpdateExpr := S.andExpr eInfo.ctx eqExpr !trsUpdateExpr
	)
      else (*-- variable is preserved --*)
	(
	  let sE = encodeVarPreserved eInfo vi in 
	    trsUpdateExpr := S.andExpr eInfo.ctx sE !trsUpdateExpr 
	)
    )
  in 
    Inthash.iter hVar compVars;
    let iE =  S.implExpr eInfo.ctx defTakenVar !trsUpdateExpr in 
    if (!debug) 
    then ( S.ppExpr eInfo.ctx iE; flush stdout); 
    S.assertExpr eInfo.ctx (
      iE
    )
      

let encodeAllUpdates gInfo eInfo cData = 
  List.iter (encodeUpdates eInfo gInfo cData.eComp.C.cVars ) cData.takenVars;
  encodeUpdateForDefault eInfo gInfo cData.defaultTakenVar cData.eComp.C.cVars


let findUpdatingTransitionsForVariable cData vi = 
  assert ( not (vi.C.isNondet));
  let foldUpdates uList (tr,vTr) = 
    if (Inthash.mem tr.C.tUpdates vi.C.vID)
    then vTr::uList
    else uList 
  in 
    List.fold_left foldUpdates [cData.defaultTakenVar] cData.takenVars



let findUpdatingTransitionsForClockFrozenVars cData vi =  
  assert (vi.C.isClock); 
  let foldUpdates uList (tr,vTr) =  
    if (C.transitionFreezesClock tr vi)
      || (C.transitionUnfreezesClock tr vi)
    then 
      vTr::uList
    else
      uList
  in 
    List.fold_left foldUpdates [] cData.takenVars


  
let encodeVariablePreservation eInfo gInfo cData vi = 
  if (vi.C.isNondet) 
  then () 
  else (
    let uList = findUpdatingTransitionsForVariable cData vi in 
    let ePres = 
      if (vi.C.isClock) 
      then encodeClockIncrement eInfo gInfo vi 
      else encodeVarPreserved eInfo vi 
    in 
    let uExpr = List.fold_left ( fun sExpr tVar -> 
				   S.andExpr eInfo.ctx sExpr (S.negateExpr eInfo.ctx tVar) 
			       ) ( S.boolConstant eInfo.ctx true ) uList 
    in 
    let iExpr = S.implExpr eInfo.ctx uExpr ePres in
      if (!debug) 
      then ( S.ppExpr eInfo.ctx iExpr; flush stdout); 
      S.assertExpr eInfo.ctx iExpr;
      if (vi.C.isClock)
      then begin
	let fList = findUpdatingTransitionsForClockFrozenVars cData vi in 
	let fExpr = List.fold_left ( fun sExpr tVar -> 
				       S.andExpr eInfo.ctx sExpr (S.negateExpr eInfo.ctx tVar) 
				   ) ( S.boolConstant eInfo.ctx true ) fList 
	in 
	let eqExpr = S.eqExpr eInfo.ctx (getClockFrozenVariableAtCurrentDepth eInfo vi) (getClockFrozenVariableAtPreviousDepth eInfo vi) in 	  
	let ifExpr = S.implExpr eInfo.ctx fExpr eqExpr in 
	  if (!debug) 
	  then ( S.ppExpr eInfo.ctx ifExpr; flush stdout); 
	  S.assertExpr eInfo.ctx ifExpr
      end
  
)

let encodePres eInfo gInfo cData = 
  let cVars = cData.eComp.C.cVars in 
    Inthash.iter (fun _ vi -> encodeVariablePreservation eInfo gInfo cData vi) cVars

(*---- Main interface functions for outside use -----*)

let initializeEncoding gInfo = 
  let eInfo = createEncodingInfo () in 
    createAllParameters gInfo eInfo ;
    createVariablesAtCurrentDepth gInfo eInfo ;
    assertInitialConditions gInfo eInfo ;
    eInfo

let increaseEncodingDepth gInfo eInfo = 
  (* 
     0. Increase depth.
     1. Create variables at current depth.
     2. For each component, 
         2.1 create component data.
         2.2 encode transition taken
         2.3 encode updates
         2.4 encode preservation 
         2.5 ...?
  *)
  let hComp comp = 
    let cData = createEncodingDataForComponent eInfo comp in 
      encodeAtLeastOneTransitionTaken eInfo cData;
      encodeAtMostOneTransitionTaken eInfo cData;
      encodeGuardImplication eInfo cData;
      encodeAllUpdates gInfo eInfo cData
	(*encodePres eInfo gInfo cData*)
  in 
    eInfo.curDepth <- eInfo.curDepth +1 ;
    createVariablesAtCurrentDepth gInfo eInfo ;
    List.iter hComp gInfo.C.gComponents 




(*-- Trace Extraction Utilities --*)



let extractAndPrintValue gInfo sModel sVar  vName  vType  = 
    match vType with 
	| C.BoolType -> 
	    let b  = S.evalBoolVar sModel sVar in 
	      Printf.printf "%s,  "  (if b then "true" else "false" );
				C.BoolConst b
	| C.RealType -> 
	    let f = S.evalRealVar sModel sVar in 
	      Printf.printf "%s, " f;
	    C.NumConst (MyUtils.convertToFloat f) 
	| C.EnumType -> 
	    let i = S.evalIntVar sModel sVar in 
	    let str = C.string_of_enum_id gInfo i in 
	      Printf.printf "%s, " str;
	    CompAst.EnumConst i
	| _ -> assert false


let extractValuesFor gInfo eInfo sModel depth _ vi = 
  let sVar = getEncodingVariableAtDepth eInfo vi depth in 
    ignore (extractAndPrintValue gInfo sModel sVar vi.C.vName vi.C.vType )
      
let extractVariableValue  ctx sModel sVar vType = 
	  match vType with 
	| C.BoolType -> 
	    let b  = S.evalBoolVar sModel sVar in 
				C.BoolConst b
	| C.RealType -> 
	    let f = S.evalRealVar sModel sVar in 
	    C.NumConst (MyUtils.convertToFloat f)
	    
	| C.EnumType -> 
	    let i = S.evalIntVar sModel sVar in 
				C.EnumConst i
	| _ -> assert false

	
let extractVariableValuesAtDepth gInfo eInfo sModel  depth = 
    Inthash.iter (extractValuesFor gInfo eInfo sModel depth) gInfo.C.gVars

let extractNondetValuesAtDepth gInfo eInfo sModel depth  = 
      Inthash.iter (extractValuesFor gInfo eInfo sModel depth) gInfo.C.gNondets

let extractClockValuesAtDepth gInfo eInfo sModel depth = 
  let hClockVar _ vi = 
    let sVar = getEncodingVariableAtDepth eInfo vi depth in 
    let fVar = getClockFrozenVariableAtDepth eInfo vi depth in 
      ignore ( 
				extractAndPrintValue gInfo sModel sVar vi.C.vName vi.C.vType;
        extractAndPrintValue gInfo sModel fVar (vi.C.vName^"_is_running") C.BoolType 
				)
  in 
    Inthash.iter hClockVar gInfo.C.gClocks


let extractTakenTransitionsAtDepth gInfo eInfo sModel depth = ()


let printHeadersForTrace gInfo = 
  (* -- 1. Print variable names --*)
  Inthash.iter (fun _ vinfo -> Printf.printf "%s, " vinfo.C.vName) gInfo.C.gVars;
  (* -- 2. Print clock names  --*)
  Inthash.iter (fun _ vinfo -> Printf.printf "%s, %s_is_running," vinfo.C.vName vinfo.C.vName ) gInfo.C.gClocks;
  (* -- 3. Print nondet names --*)
  Inthash.iter (fun _ vinfo -> Printf.printf "%s, " vinfo.C.vName) gInfo.C.gNondets;
  Printf.printf "\n"


let extractValuesAtDepth gInfo eInfo  =
  let ctx = eInfo.ctx in 
  let depthLim = eInfo.curDepth in 
  match (S.check ctx)
  with 
    | S.SAT -> (
	let sModel =  S.getModel ctx  in
	printHeadersForTrace gInfo;
	for depth = 0 to depthLim do
	  extractVariableValuesAtDepth gInfo eInfo sModel depth;
	  extractClockValuesAtDepth gInfo eInfo sModel depth;
	  extractNondetValuesAtDepth gInfo eInfo sModel depth;
	  extractTakenTransitionsAtDepth gInfo eInfo sModel depth;
	  Printf.printf "\n"
	done
    )
    | S.UNSAT  -> Printf.fprintf stderr "UNSAT \n"; assert false 
    | S.FAIL -> Printf.fprintf stderr "FAIL \n"; assert false
      

    



(*-- what else? --*)
  
exception PropertyFailed
exception AllPropsResolved

let chkExpr gInfo eInfo e1 eVar e2 = 
  let sExpr = encodeExpression eInfo e1 in
  let cExpr = encodeExpression eInfo e2 in 
    S.push eInfo.ctx;
    S.assertExpr eInfo.ctx sExpr;
    S.assertExpr eInfo.ctx cExpr;
    match (S.check eInfo.ctx) with 
      | S.SAT -> (
	  Printf.printf "\n ------FOUND PROPERTY ------- \n";
	  XmlAstPrinter.xmlPrintExpr stdout e1;
	  Printf.printf "\n -- AND -- \n";
	  XmlAstPrinter.xmlPrintExpr stdout e2;
	  let sModel =  S.getModel eInfo.ctx  in
	  let actualValue = S.evalRealVar sModel eVar in 
	  Printf.printf "TRACE: \n";
	  extractValuesAtDepth gInfo eInfo;
	  Printf.printf "\n";
	  flush stdout;
	  S.pop eInfo.ctx;
	  (true, (float_of_string actualValue))
	)
      | S.UNSAT -> S.pop eInfo.ctx; 
	(false, 0.0);
      | S.FAIL -> assert false

let checkPropertyViolation gInfo eInfo = 
  let someProp = ref false in 
  let hProp p =
    if not (p.C.pViolated)
    then ( 
      let e = p.C.pExpr in 
      let sE = encodeExpression eInfo e in 
      let seNeg = S.negateExpr eInfo.ctx sE in 
      S.push eInfo.ctx;
      S.assertExpr eInfo.ctx seNeg;
      match (S.check eInfo.ctx) 
      with 
	| S.SAT -> (
	  Printf.printf "\n ------FOUND PROPERTY VIOLATION------- \n";
	  XmlAstPrinter.xmlPrintExpr stdout p.C.pExpr;
	  Printf.printf "TRACE: \n";
	  extractValuesAtDepth gInfo eInfo;
	  Printf.printf "\n";
	  S.pop eInfo.ctx;
	  p.C.pViolated <- true
	)
	| S.UNSAT -> ( someProp := true; S.pop eInfo.ctx )
	| S.FAIL -> assert false
    )
  in 
    List.iter hProp gInfo.C.gProps ;
    if (not (!someProp)) && ((List.length gInfo.C.gMax) = 0)
    then raise AllPropsResolved 


(*-- handle max objective --*)

exception LimExceeded

let checkNextMaxObj gInfo eInfo mObj vExpr = 
  if (mObj.C.mCur >= mObj.C.mHi)
  then raise LimExceeded 
  else (
    let tryVal = 
      if (mObj.C.mCur < mObj.C.mLo) 
      then mObj.C.mLo
      else mObj.C.mCur +. mObj.C.mStep
    in 
      (*-- check if we can get the trace --*)
    let e = C.ARelOp (C.AGt, mObj.C.mExpr, (C.AConst (C.NumConst tryVal)) ) in
    let (b,aVal) = chkExpr gInfo eInfo e vExpr mObj.C.mCond in 
    if (b)
    then (
      mObj.C.mCur <- aVal; 
      true )
    else false
  )


let handleMaxObj gInfo eInfo mObj = 
  try 
    if not (mObj.C.limExceeded) 
    then  (
      Printf.printf " Maxing expression: \n";
      XmlAstPrinter.xmlPrintExpr stdout mObj.C.mExpr;
      Printf.printf  " \t trying %f <--- \n" mObj.C.mCur ;
      let vName = "tmpMax"^(string_of_int eInfo.curDepth) in 
      let vExpr = S.createVariable eInfo.ctx vName C.RealType in 
      let eExpr = encodeExpression eInfo mObj.C.mExpr in 
      let veExpr = S.eqExpr eInfo.ctx vExpr eExpr in 
      S.assertExpr eInfo.ctx veExpr;
      while (checkNextMaxObj gInfo eInfo mObj vExpr)
      do
	Printf.printf  "\t Current bound is: %f \n" mObj.C.mCur 
      done
    )
  with | LimExceeded ->  (
    Printf.printf "Limit Exceeded @ %f \n" mObj.C.mCur;
    mObj.C.limExceeded <- true
  )

let handleMaximization eInfo gInfo = 
  List.iter (handleMaxObj gInfo eInfo) gInfo.C.gMax
	
	
	
	
