(*--- 
  Printing the AST in XML 
  ---*)

open MyUtils

module C= CompAst

let indentLevel = ref 0

let indent oc = 
  for i = 0 to !indentLevel
  do
    ignore (Printf.fprintf oc  "   ")
  done

let xmlOpen oc str = 
  incr indentLevel;
  indent oc ;
  ignore ( Printf.fprintf oc "<%s>\n" str )

let xmlClose oc str = 
  indent oc;
  ignore ( Printf.fprintf oc "<%s\\>\n" str );
  decr indentLevel

let xmlPrintProp oc str1 str2 = 
  indent oc;
  ignore (
    Printf.fprintf oc "< %s = %s \\> \n" str1 str2
  )

let xmlPrintTypeInfo oc ty = 
  match ty with 
    | C.RealType -> xmlPrintProp oc  "type" "real"
    | C.StringType -> xmlPrintProp oc  "type" "string"
    | C.BoolType -> xmlPrintProp oc  "type" "bool"
    | C.EnumType ->   xmlPrintProp oc  "type" "enum"
    | _ -> assert false

let xmlPrintRangeInfo oc rSpec = 
  match rSpec with 
    | C.RangeSpecNone -> xmlPrintProp oc "range" "none"
    | C.RangeSpecInterval (f1,f2) -> xmlPrintProp oc "range" (Printf.sprintf " IN [%f,%f]" f1 f2)
    | C.RangeSpecEnum iList -> xmlPrintProp oc "range" (" IN " ^ ( String.concat "," (List.map string_of_int iList)))
    | C.RangeSpecEq c -> xmlPrintProp oc "range" ("EQ "^(C.string_of_constinfo c))
    | _ -> assert false

let xmlPrintVarinfo oc vi = 
  xmlOpen oc "varInfo";
  xmlPrintProp oc "name" vi.C.vName;
  xmlPrintProp oc "id" (string_of_int vi.C.vID);
  xmlPrintTypeInfo oc vi.C.vType;
  xmlPrintRangeInfo oc vi.C.vRange;
  if (vi.C.isNondet)
  then xmlPrintProp oc "nondet" "true";
  if (vi.C.isClock)
  then xmlPrintProp oc "clock" "true";
  xmlClose oc "varInfo"


let rec xmlPrintExpr oc e = 
  xmlOpen oc "expr";
  begin
    match e with 
      | C.ABoolOp (b,e1,e2) -> 
	(
	  xmlPrintProp oc "boolOp" (C.string_of_boolOp b );
	  xmlPrintExpr oc e1;
	  xmlPrintExpr oc e2
	)
      | C.ANot e1 -> 
	(
	  xmlPrintProp oc "boolOp" "not";
	  xmlPrintExpr oc e1
	)

      | C.ARelOp (r,e1,e2) -> 
	(
	  xmlPrintProp oc "relOp" (C.string_of_relOp r) ;
	  xmlPrintExpr oc e1;
	  xmlPrintExpr oc e2
	)
      | C.AArithOp (b,e1,e2) ->
	(
	  xmlPrintProp oc "arithOp"  (C.string_of_arithOp b);
	  xmlPrintExpr oc e1;
	  xmlPrintExpr oc e2
	)
      | C.ANeg e1 ->  (
	xmlPrintProp oc "arithOp" "UNARY MINUS";
	xmlPrintExpr oc e1
      )
      | C.AVar vi ->  (
	xmlPrintProp oc "variable" vi.C.vName
      )
      | C.AConst ci ->  (
	xmlPrintProp oc "constant" 
	  (
	    C.string_of_constinfo ci
	  )
      )
      | C.ADefined pi -> xmlPrintProp oc "defined" pi.C.pName 
  end;
  xmlClose oc "expr"

let xmlPrintParams oc pi = 
  xmlOpen oc "paramInfo";
  xmlPrintProp oc "name" pi.C.pName;
  xmlPrintProp oc "id" (string_of_int pi.C.pID);
  xmlPrintTypeInfo oc pi.C.pType;
  xmlPrintRangeInfo oc pi.C.pRange;
  if (pi.C.pIsUserInp)
  then xmlPrintProp oc "userinput" "true";
  xmlClose oc "paramInfo"


let xmlPrintTransition oc cTrs = 
  xmlOpen oc "transition";
  xmlOpen oc "guard";
  xmlPrintExpr oc cTrs.C.tGuard;
  xmlClose oc "guard";
  xmlOpen oc "updates";
  Inthash.iter (fun i e -> ( 
    xmlPrintProp oc "lhsVarID" (string_of_int i);
    xmlPrintExpr oc e
  )) cTrs.C.tUpdates; 
  xmlClose oc "updates";
  xmlOpen oc "freezeClock";
  List.iter (fun vi -> xmlPrintProp oc "cName" vi.C.vName) cTrs.C.tClockFreeze;
  xmlClose oc "freezeClock";
  xmlOpen oc "unfreezeClock";
  List.iter (fun vi -> xmlPrintProp oc "cName" vi.C.vName) cTrs.C.tClockUnfreeze;
  xmlClose oc "unfreezeClock";
  xmlClose oc "transition"

let xmlPrintCompInfo oc cI = 
  xmlOpen oc "component";
  xmlPrintProp oc "name" (cI.C.cName^"::"^(string_of_int cI.C.cID));
  xmlOpen oc "vars";
  Inthash.iter (fun _ pi -> xmlPrintVarinfo oc pi ) cI.C.cVars;
  xmlClose oc "vars";
  xmlPrintExpr oc cI.C.cInit;
  List.iter (xmlPrintTransition oc) cI.C.cTransitions;
  xmlClose oc "component"

let xmlPrintGlobinfo oc gInfo = 
  xmlOpen oc "global";

  xmlOpen oc "vars";
  Inthash.iter (fun _ vi -> xmlPrintVarinfo oc vi ) gInfo.C.gVars;
  xmlClose oc "vars";

  xmlOpen oc "params";
  Inthash.iter (fun _ pi -> xmlPrintParams oc pi ) gInfo.C.gParams;
  xmlClose oc "params";
  
  xmlOpen oc "clocks";
  Inthash.iter (fun _ vi -> xmlPrintVarinfo oc vi ) gInfo.C.gClocks;
  xmlClose oc "clocks";

  xmlOpen oc "nondets";
  Inthash.iter (fun _ vi -> xmlPrintVarinfo oc vi ) gInfo.C.gNondets;
  xmlClose oc "nondets";
  
  xmlOpen oc "enumSymbols";
  Inthash.iter (fun i str -> xmlPrintProp oc "enumSym" (str^":"^(string_of_int i)) ) gInfo.C.gEnums;
  xmlClose oc "enumSymbols";

  xmlOpen oc "asserts";
  List.iter (fun p -> xmlPrintExpr oc p.C.pExpr) gInfo.C.gProps;
  xmlClose oc "asserts";

  xmlOpen oc "assumes";
  List.iter (fun e -> xmlPrintExpr oc e) gInfo.C.gAssumes;
  xmlClose oc "assumes";

  
  xmlOpen oc "components";
  List.iter (fun c -> xmlPrintCompInfo oc c) gInfo.C.gComponents;
  xmlClose oc "components";
  
  xmlClose oc "global"
