/*-- 

  Interface from commonly used yices SMT solver functinality to
  OCAML. The idea is to support bounded model checking of access net
  models directly from the OCaml code.
 
  Author: Sriram Sankaranarayanan <srirams@colorado.edu>
        
  ---*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
//extern "C" {
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/fail.h>
//}
#include "yices_c.h"

#define YICES_CONTEXT(a)  get_yices_context(a)
#define YICES_EXPR(a)  get_yices_expr(a)
#define ASSERTION_ID(a) (assertion_id) (Int_val(a))
#define YICES_MODEL(a) get_yices_model(a)

/*-- convert a lbool to value --*/
value  Val_lbool( lbool retVal){
  if (retVal == l_false) 
    return (Val_int(0));
  if (retVal == l_true)
    return (Val_int(1));
  /*-- l_undef  --> -1 --*/
  return (Val_int(-1));
}


yices_context get_yices_context( value a){
  return (yices_context) Field(a,0);
}

yices_expr get_yices_expr( value a){
  return ((yices_expr) Field(a,0));
}

yices_model get_yices_model(value a){
  return (yices_model) Field(a,0);
}

value Val_expr(yices_expr a){
  value rVal = alloc_tuple(1);
  Store_field(rVal,0,(value) a);
  return rVal;
}





/*--- 
  
  Interface functions to create, delete, add assertions and 
  check satisfiability of a context 

  ---*/

CAMLprim value c_ex_yices_mk_context(){
  CAMLparam0();
  yices_context c = yices_mk_context();
  value retVal = alloc_tuple(1);
  Store_field(retVal, 0, (value) c);
  
  CAMLreturn(retVal); /*-- just convert c into a value --*/
}

CAMLprim value c_ex_yices_del_context( value v1){
  CAMLparam1(v1);
  yices_context ctx = YICES_CONTEXT(v1);
  yices_del_context(ctx);
  CAMLreturn (Val_unit);
}

CAMLprim value c_ex_yices_reset( value v1){
  CAMLparam1(v1);
  yices_context ctx = YICES_CONTEXT( v1);
  yices_reset(ctx);
  CAMLreturn (Val_unit);
}

CAMLprim value c_ex_yices_dump_context(value v1){
  CAMLparam1(v1);
  yices_context ctx = YICES_CONTEXT(v1);
  yices_dump_context(ctx);
  CAMLreturn (Val_unit);
}


CAMLprim value c_ex_yices_push(value v1){
  CAMLparam1(v1);
  yices_push(YICES_CONTEXT(v1));
  CAMLreturn (Val_unit);
}


CAMLprim value c_ex_yices_pop(value v1){
  CAMLparam1(v1);
  yices_pop( YICES_CONTEXT(v1));
  CAMLreturn (Val_unit);
}

CAMLprim value c_ex_yices_assert(value vCtx, value vExpr){
  CAMLparam2(vCtx,vExpr);
  yices_assert( YICES_CONTEXT(vCtx), YICES_EXPR(vExpr));
  CAMLreturn (Val_unit);
}

CAMLprim value c_ex_yices_assert_retractable(value vCtx, value vExpr){
  CAMLparam2(vCtx, vExpr);
  assertion_id id = yices_assert_retractable(YICES_CONTEXT(vCtx), 
					     YICES_EXPR(vExpr));
  CAMLreturn(Val_int(id));
}

CAMLprim value c_ex_yices_retract(value vCtx, value vId){
  CAMLparam2(vCtx, vId);
  yices_retract(YICES_CONTEXT(vCtx), ASSERTION_ID(vId));
  CAMLreturn (Val_unit);
}

CAMLprim value c_ex_yices_inconsistent(value vCtx){
  CAMLparam1(vCtx);
  int retVal = yices_inconsistent(YICES_CONTEXT(vCtx));
  if (retVal)
    CAMLreturn (Val_true);
  
  CAMLreturn (Val_false);
}

CAMLprim value c_ex_yices_check(value vCtx){
  CAMLparam1(vCtx);
  lbool retVal = yices_check(YICES_CONTEXT(vCtx));

  if (retVal == l_undef){
    printf("Warning: check returns undefined return value. \n");
  }

  CAMLreturn (Val_lbool(retVal));
}

CAMLprim value c_ex_yices_get_model(value vCtx){
  CAMLparam1(vCtx);
  yices_model m = yices_get_model(YICES_CONTEXT(vCtx));
  if (m == 0){
    printf ("Warning: null model returned in c_ex_yices_get_model\n");
  }
  value retVal = alloc_tuple(1);
  Store_field(retVal, 0, (value) m);
  
  CAMLreturn (retVal); /*-- just convert m into a value --*/
}


/*--- 
  Interface functions to build expressions 
  ---*/


CAMLprim value c_ex_yices_mk_true (value vCtx){
  CAMLparam1(vCtx);
  yices_expr ex =  yices_mk_true(YICES_CONTEXT(vCtx));
  CAMLreturn (Val_expr(ex));
}

CAMLprim value c_ex_yices_mk_false (value vCtx){
  CAMLparam1(vCtx);
  yices_expr ex =  yices_mk_false(YICES_CONTEXT(vCtx));
  CAMLreturn (Val_expr(ex));
}

CAMLprim value c_ex_yices_mk_bool_var(value vCtx, value vName){
  CAMLparam2(vCtx, vName);
  yices_var_decl vDecl = yices_mk_bool_var_decl(YICES_CONTEXT(vCtx),String_val(vName));
  yices_expr ex = yices_mk_bool_var_from_decl( YICES_CONTEXT(vCtx), vDecl);
  CAMLreturn (Val_expr(ex));
}

CAMLprim value c_ex_yices_mk_fresh_bool_var(value vCtx){
  CAMLparam1(vCtx);
  yices_expr ex = yices_mk_fresh_bool_var( YICES_CONTEXT(vCtx));
  CAMLreturn (Val_expr(ex));
}

/* yices_expr 	yices_mk_or (yices_context ctx, yices_expr args[], unsigned n) */
/* Warning this code has been cut and paste. Remember to propagate bug fixes to 
   yices_mk_and function below */
CAMLprim value c_ex_yices_mk_or(value vCtx, value vArgs){
  CAMLparam2(vCtx, vArgs);
  unsigned n = Wosize_val(vArgs);
  yices_expr ex;
  if (n== 0) {
    ex =  yices_mk_true(YICES_CONTEXT(vCtx));
  } else {
    yices_expr * args = (yices_expr*) malloc(n * sizeof(yices_expr));
    unsigned i;
    for (i=0; i < n; ++i){
      args[i] = YICES_EXPR(Field(vArgs,i));
    }
    
    ex= yices_mk_or( YICES_CONTEXT(vCtx) , args, n);
    free(args); // SS: Is this SAFE? need to test.
    
  }
  CAMLreturn (Val_expr(ex));
}


/* yices_expr 	yices_mk_or_2 (yices_context ctx, yices_expr args[], unsigned n) */
/* Warning this code has been cut and paste. Remember to propagate bug fixes to 
   yices_mk_and function below */
CAMLprim value c_ex_yices_mk_or_2(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);

  yices_expr  args[2]; 
  args[0] = YICES_EXPR(vArg1);
  args[1] = YICES_EXPR(vArg2);
  yices_expr ex = yices_mk_or( YICES_CONTEXT(vCtx) , args, 2);
  CAMLreturn (Val_expr(ex));
}



/* yices_expr 	yices_mk_and_2 (yices_context ctx, yices_expr args[], unsigned n) */
/* Warning this code has been cut and paste. Remember to propagate bug fixes to 
   yices_mk_and function below */
CAMLprim value c_ex_yices_mk_and_2(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);

  yices_expr  args[2]; 
  args[0] = YICES_EXPR(vArg1);
  args[1] = YICES_EXPR(vArg2);
  yices_expr ex = yices_mk_and( YICES_CONTEXT(vCtx) , args, 2);
  CAMLreturn (Val_expr(ex));
}


/* yices_expr 	yices_mk_and (yices_context ctx, yices_expr args[], unsigned n) */
/*SS: Warning: Cut and Paste from yices_mk_or above */
CAMLprim value c_ex_yices_mk_and(value vCtx, value vArgs){
  CAMLparam2(vCtx, vArgs);
  unsigned n = Wosize_val(vArgs);
  //assert( n > 0);
  yices_expr ex;

  if (n > 0 ){
    yices_expr * args = (yices_expr*) malloc(n * sizeof(yices_expr));
    unsigned i;
    for (i=0; i < n; ++i){
      args[i] = YICES_EXPR(Field(vArgs,i));
    }
    
    ex = yices_mk_and( YICES_CONTEXT(vCtx) , args, n);
    free(args); // SS: Is this SAFE? need to test.
  } else 
    ex = yices_mk_true(YICES_CONTEXT(vCtx));

  CAMLreturn (Val_expr(ex));
}

/*--
yices_expr 	yices_mk_eq (yices_context ctx, yices_expr a1, yices_expr a2)
 	CAMLreturn an expression representing a1 = a2.
	--*/

CAMLprim value c_ex_yices_mk_eq( value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);
  CAMLreturn (Val_expr (
		   yices_mk_eq( YICES_CONTEXT(vCtx), YICES_EXPR(vArg1), YICES_EXPR(vArg2) )
			));
  
}
/*--
yices_expr 	yices_mk_diseq (yices_context ctx, yices_expr a1, yices_expr a2)
 	CAMLreturn an expression representing a1 /= a2.
	--*/
CAMLprim value c_ex_yices_mk_diseq(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);
  CAMLreturn (Val_expr (
		   yices_mk_diseq( YICES_CONTEXT(vCtx), YICES_EXPR(vArg1), YICES_EXPR(vArg2) )
			));
}

/*--
yices_expr 	yices_mk_lt (yices_context ctx, yices_expr a1, yices_expr a2)
 	CAMLreturn an expression representing a1 < a2.
	--*/
CAMLprim value c_ex_yices_mk_lt(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);
  CAMLreturn (Val_expr (
		   yices_mk_lt( YICES_CONTEXT(vCtx), YICES_EXPR(vArg1), YICES_EXPR(vArg2) )
			));
}
/*--
yices_expr 	yices_mk_le (yices_context ctx, yices_expr a1, yices_expr a2)
 	CAMLreturn an expression representing a1 <= a2.
	--*/
CAMLprim value c_ex_yices_mk_le(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);
  CAMLreturn (Val_expr (
		   yices_mk_le( YICES_CONTEXT(vCtx), YICES_EXPR(vArg1), YICES_EXPR(vArg2) )
			));
}
/*--
yices_expr 	yices_mk_gt (yices_context ctx, yices_expr a1, yices_expr a2)
 	CAMLreturn an expression representing a1 > a2.
	--*/
CAMLprim value c_ex_yices_mk_gt(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);
  CAMLreturn (Val_expr (
		   yices_mk_gt( YICES_CONTEXT(vCtx), YICES_EXPR(vArg1), YICES_EXPR(vArg2) )
			));
}

/*--
yices_expr 	yices_mk_ge (yices_context ctx, yices_expr a1, yices_expr a2)
 	CAMLreturn an expression representing a1 >= a2. 
	--*/
CAMLprim value c_ex_yices_mk_ge(value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx, vArg1, vArg2);
  CAMLreturn (Val_expr (
		   yices_mk_ge( YICES_CONTEXT(vCtx), YICES_EXPR(vArg1), YICES_EXPR(vArg2) )
			));
}


/*--
yices_expr 	yices_mk_ite (yices_context ctx, yices_expr c, yices_expr t, yices_expr e)
 	CAMLreturn an expression representing (if c t e).
	--*/

CAMLprim value c_ex_yices_mk_ite(value vCtx, value vCond, value vThen, value vElse){
  CAMLparam4(vCtx, vCond, vThen, vElse);
  CAMLreturn (Val_expr (
		   yices_mk_ite( YICES_CONTEXT(vCtx), YICES_EXPR(vCond), YICES_EXPR(vThen), YICES_EXPR(vElse) )
			));
}

/*--
yices_expr 	yices_mk_not (yices_context ctx, yices_expr a)
 	CAMLreturn an expression representing (not a). 
	--*/

CAMLprim value c_ex_yices_mk_not(value vCtx, value vCond){
  CAMLparam2(vCtx, vCond);
  CAMLreturn (Val_expr (
		   yices_mk_not( YICES_CONTEXT(vCtx), YICES_EXPR(vCond) )
			));
}

/*--
void 	yices_pp_expr (yices_expr e)
 	Pretty print the given expression in the standard output. 
	--*/

CAMLprim value c_ex_yices_pp_expr( value vExpr){
  CAMLparam1(vExpr);
  yices_pp_expr(YICES_EXPR(vExpr));
  printf ("\n");
  fflush(stdout);
  CAMLreturn (Val_unit);
}


/*---
 
yices_expr 	yices_mk_sum (yices_context ctx, yices_expr args[], unsigned n)
 	CAMLreturn an expression representing args[0] + ... + args[n-1].
yices_expr 	yices_mk_sub (yices_context ctx, yices_expr args[], unsigned n)
 	CAMLreturn an expression representing args[0] - ... - args[n-1].
yices_expr 	yices_mk_mul (yices_context ctx, yices_expr args[], unsigned n)
 	CAMLreturn an expression representing args[0] * ... * args[n-1]. 

--*/

CAMLprim value c_ex_yices_add( value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx,vArg1,vArg2);
  yices_expr args[2];
  args[0] = YICES_EXPR(vArg1);
  args[1] = YICES_EXPR(vArg2);
  CAMLreturn (Val_expr( yices_mk_sum(YICES_CONTEXT(vCtx), args, 2) ));
}


CAMLprim value c_ex_yices_sub( value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx,vArg1,vArg2);
  yices_expr args[2];
  args[0] = YICES_EXPR(vArg1);
  args[1] = YICES_EXPR(vArg2);
  CAMLreturn(Val_expr( yices_mk_sub(YICES_CONTEXT(vCtx), args, 2) ));
}

CAMLprim value c_ex_yices_mul( value vCtx, value vArg1, value vArg2){
  CAMLparam3(vCtx,vArg1,vArg2);
  yices_expr args[2];
  args[0] = YICES_EXPR(vArg1);
  args[1] = YICES_EXPR(vArg2);
  CAMLreturn(Val_expr( yices_mk_mul(YICES_CONTEXT(vCtx), args, 2) ));
}

/*
yices_expr 	yices_mk_num (yices_context ctx, int n)
 	CAMLreturn an expression representing the given integer. 
*/

CAMLprim value c_ex_yices_mk_num_from_str(value vCtx, value vStr){
  CAMLparam2(vCtx,vStr);
  CAMLreturn( Val_expr ( yices_mk_num_from_string( YICES_CONTEXT(vCtx), String_val(vStr)) ));
}

CAMLprim value c_ex_yices_mk_num(value vCtx, value vArg){
  CAMLparam2(vCtx,vArg);
  CAMLreturn(Val_expr (yices_mk_num( YICES_CONTEXT(vCtx), Int_val(vArg)) ));
}

CAMLprim value c_ex_yices_mk_int_var(value vCtx, value vName){
  CAMLparam2(vCtx,vName);
  yices_context ctx = YICES_CONTEXT(vCtx);
  yices_type intType = yices_mk_type(ctx,"int");
  yices_var_decl var = yices_mk_var_decl(ctx, String_val(vName), intType);
  CAMLreturn(Val_expr (yices_mk_var_from_decl(ctx,var)));
}



CAMLprim value c_ex_yices_mk_real_var(value vCtx, value vName){
  CAMLparam2(vCtx,vName);
  yices_context ctx = YICES_CONTEXT(vCtx);
  yices_type intType = yices_mk_type(ctx,"real");
  yices_var_decl var = yices_mk_var_decl(ctx, String_val(vName), intType);
  CAMLreturn (Val_expr (yices_mk_var_from_decl(ctx,var)));
}



CAMLprim value c_ex_yices_eval_bool_var(value vModel, value vExpr){
  CAMLparam2(vModel,vExpr);
  lbool retVal = yices_evaluate_in_model(YICES_MODEL(vModel), YICES_EXPR(vExpr));
  CAMLreturn( Val_lbool(retVal));
}


CAMLprim value c_ex_yices_eval_real_var(value vModel, value vExpr){
  CAMLparam2(vModel,vExpr);
  value rValue = alloc(1,Double_tag);
  yices_var_decl var= yices_get_var_decl( YICES_EXPR(vExpr) );
  double retValue = 1982.1982;
  int stat = yices_get_double_value(YICES_MODEL(vModel),var, &retValue );
  if (stat == 0){
    fprintf(stderr, "error: eval_real_var yicesOcamlInterface.c \n");
  }
  Double_val(rValue) = retValue;
  CAMLreturn( rValue);
}

CAMLprim value c_ex_yices_eval_int_var (value vModel, value vExpr){
  CAMLparam2(vModel,vExpr);
  yices_var_decl var= yices_get_var_decl( YICES_EXPR(vExpr) );
  int retValue;
  int stat = yices_get_int_value(YICES_MODEL(vModel),var, &retValue );
  if (stat == 0){
    fprintf(stderr, "error: eval_int_var yicesOcamlInterface.c \n");
  }

  CAMLreturn( Val_int(retValue));
}

